"""Exceptions for the OVO Energy API."""


# Base Exceptions
class OVOEnergyException(Exception):
    """Base exception for OVO Energy."""


class OVOEnergyNoAccount(OVOEnergyException):
    """Exception for no account found."""


# API Exceptions
class OVOEnergyAPIException(OVOEnergyException):
    """Exception for API exceptions."""


class OVOEnergyAPINotAuthorized(OVOEnergyAPIException):
    """Exception for API client not authorized."""


class OVOEnergyAPINoCookies(OVOEnergyAPIException):
    """Exception for no cookies found."""



"""Setup for tests."""

from typing import Final

USERNAME: Final[str] = "test"
PASSWORD: Final[str] = "test"
ACCOUNT: Final[int] = 123456789
ACCOUNT_BAD: Final[int] = 654321789

RESPONSE_JSON_BASIC: Final[dict] = {"test": "test"}

RESPONSE_JSON_AUTH: Final[dict] = {"code": "test"}

RESPONSE_JSON_TOKEN: Final[dict] = {
    "accessToken": {"value": "test"},
    "expiresIn": 3600,
    "refreshExpiresIn": 0,
}

RESPONSE_JSON_BOOTSTRAP_ACCOUNTS: Final[dict] = {
    "accountIds": [ACCOUNT],
    "customerId": "5cafe9c4-a942-46b5-a67c-5882eba0a03c",
    "selectedAccountId": ACCOUNT,
    "accounts": [
        {
            "accountId": ACCOUNT,
            "isPayg": False,
            "isBlocked": False,
            "supplies": [
                {
                    "mpxn": "3456766576",
                    "fuel": "gas",
                    "isOnboarding": False,
                    "start": "2024-01-01T23:00:00Z",
                    "isPayg": False,
                    "supplyPointInfo": {
                        "meterType": "AB123",
                        "meterNotFound": False,
                        "address": ["ADDR"],
                    },
                },
                {
                    "mpxn": "4536756746",
                    "fuel": "electricity",
                    "isOnboarding": False,
                    "start": "2024-01-01T23:00:00Z",
                    "isPayg": False,
                    "supplyPointInfo": {
                        "meterType": "AB123",
                        "meterNotFound": False,
                        "address": ["ADDR"],
                    },
                },
            ],
        }
    ],
    "isFirstLogin": False,
}

RESPONSE_JSON_DAILY_USAGE: Final[dict] = {
    "electricity": [
        {
            "consumption": 10.24,
            "interval": {
                "start": "2024-01-01T00:00:00Z",
                "end": "2024-01-01T23:59:59.999000Z",
            },
            "meterReadings": None,
            "hasHalfHourData": None,
            "cost": {"amount": "2.94", "currencyUnit": "GBP"},
        }
    ],
    "gas": [
        {
            "consumption": 14.68,
            "volume": None,
            "interval": {
                "start": "2024-01-01T00:00:00Z",
                "end": "2024-01-01T23:59:59.999000Z",
            },
            "meterReadings": None,
            "hasHalfHourData": None,
            "cost": {"amount": "2.56", "currencyUnit": "GBP"},
        },
    ],
}

RESPONSE_JSON_PLANS: Final[dict] = {
    "electricity": [],
    "gas": [],
}

RESPONSE_JSON_FOOTPRINT: Final[dict] = {
    "from": "2024-01-01T00:00:00Z",
    "to": "2024-01-01T23:59:59.999000Z",
    "carbonReductionProductIds": [],
    "carbonFootprint": {
        "carbonKg": 2200.1234,
        "carbonSavedKg": 0.0,
        "kWh": 1578.3246,
        "breakdown": {
            "electricity": {
                "carbonKg": 200.1234,
                "carbonSavedKg": 230.02,
                "kWh": 65645.92,
            },
            "gas": {
                "carbonKg": 2000.1234,
                "carbonSavedKg": 340.02,
                "kWh": 10664.74363579,
            },
        },
    },
}

RESPONSE_JSON_INTENSITY: Final[dict] = {
    "forecast": [
        {
            "from": "2pm",
            "intensity": 82,
            "level": "low",
            "colour": "#0A9928",
            "colourV2": "#0D8426",
        },
    ],
    "current": "low",
    "greentime": None,
}



"""Fixtures for testing."""

from collections.abc import AsyncGenerator

from aiohttp import ClientSession
from aioresponses import aioresponses
import pytest

from ovoenergy import OVOEnergy

from . import (
    ACCOUNT,
    RESPONSE_JSON_AUTH,
    RESPONSE_JSON_BOOTSTRAP_ACCOUNTS,
    RESPONSE_JSON_DAILY_USAGE,
    RESPONSE_JSON_FOOTPRINT,
    RESPONSE_JSON_INTENSITY,
    RESPONSE_JSON_PLANS,
    RESPONSE_JSON_TOKEN,
)


@pytest.fixture(autouse=True)
def mock_aioresponse():
    """Return a client session."""
    with aioresponses() as mocker:
        mocker.post(
            "https://my.ovoenergy.com/api/v2/auth/login",
            payload=RESPONSE_JSON_AUTH,
            status=200,
            repeat=True,
        )
        mocker.get(
            "https://my.ovoenergy.com/api/v2/auth/token",
            payload=RESPONSE_JSON_TOKEN,
            status=200,
            repeat=True,
        )
        mocker.get(
            "https://smartpaymapi.ovoenergy.com/first-login/api/bootstrap/v2/",
            payload=RESPONSE_JSON_BOOTSTRAP_ACCOUNTS,
            status=200,
            repeat=True,
        )
        mocker.get(
            f"https://smartpaymapi.ovoenergy.com/usage/api/daily/{ACCOUNT}?date=2024-01",
            payload=RESPONSE_JSON_DAILY_USAGE,
            status=200,
            repeat=True,
        )
        mocker.get(
            f"https://smartpaymapi.ovoenergy.com/usage/api/half-hourly/{ACCOUNT}?date=2024-01-01",
            payload=RESPONSE_JSON_DAILY_USAGE,
            status=200,
            repeat=True,
        )
        mocker.get(
            f"https://smartpaymapi.ovoenergy.com/orex/api/plans/{ACCOUNT}",
            payload=RESPONSE_JSON_PLANS,
            status=200,
            repeat=True,
        )
        mocker.get(
            f"https://smartpaymapi.ovoenergy.com/carbon-api/{ACCOUNT}/footprint",
            payload=RESPONSE_JSON_FOOTPRINT,
            status=200,
            repeat=True,
        )
        mocker.get(
            "https://smartpaymapi.ovoenergy.com/carbon-bff/carbonintensity",
            payload=RESPONSE_JSON_INTENSITY,
            status=200,
            repeat=True,
        )

        yield mocker


@pytest.fixture
async def ovoenergy_client() -> AsyncGenerator[OVOEnergy, None]:
    """Return a OVOEnergy client."""
    async with ClientSession() as session:
        yield OVOEnergy(client_session=session)



"""Models."""

from dataclasses import dataclass
from datetime import datetime


@dataclass
class OVOInterval:
    """Interval model."""

    start: datetime
    end: datetime


@dataclass
class OVOMeterReadings:
    """Meter readings model."""

    start: float
    end: float


@dataclass
class OVOCost:
    """Cost model."""

    amount: float | None
    currency_unit: str | None


@dataclass
class OVODailyElectricity:
    """Daily electricity model."""

    consumption: float | None
    interval: OVOInterval | None
    meter_readings: OVOMeterReadings | None
    has_half_hour_data: bool | None
    cost: OVOCost | None


@dataclass
class OVODailyGas:
    """Daily gas model."""

    consumption: float | None
    volume: float | None
    interval: OVOInterval | None
    meter_readings: OVOMeterReadings | None
    has_half_hour_data: bool | None
    cost: OVOCost | None


@dataclass
class OVOHalfHour:
    """Half hour model."""

    consumption: float
    interval: OVOInterval
    unit: str


@dataclass
class OVODailyUsage:
    """Daily usage model."""

    electricity: list[OVODailyElectricity] | None
    gas: list[OVODailyGas] | None


@dataclass
class OVOHalfHourUsage:
    """Half hour usage model."""

    electricity: list[OVOHalfHour] | None
    gas: list[OVOHalfHour] | None


@dataclass
class OVOPlan:
    """Plan model."""

    standing_charge: float | None
    unit_rate: float | None
    tariff: str | None



"""OAuth model."""

from dataclasses import dataclass
from datetime import datetime


@dataclass
class OAuth:
    """OAuth model."""

    access_token: str
    expires_in: int
    refresh_expires_in: int
    expires_at: datetime



"""
Provides ovoenergy version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update ovoenergy` to change this file.

from incremental import Version

__version__ = Version("ovoenergy", 2, 0, 1, dev=0)
__all__ = ["__version__"]



"""Plan Models."""

from dataclasses import dataclass
from typing import Any


@dataclass
class OVOPlanRate:
    """Plan rate model."""

    amount: float
    currency_unit: str


@dataclass
class OVOPlanStatus:
    """Plan status model."""

    active: bool
    in_renewal: bool
    in_loss: bool
    loss_complete: bool
    has_future_contracts: bool


@dataclass
class OVOPlanUnitRate:
    """Unit rate model."""

    name: str
    unit_rate: OVOPlanRate


@dataclass
class OVOPlanElectricity:
    """Plan electricity model."""

    name: str
    exit_fee: OVOPlanRate
    contract_start_date: str
    contract_end_date: Any
    contract_type: str
    is_in_renewal: bool
    has_future_contracts: bool
    mpxn: str
    msn: str
    personal_projection: float
    standing_charge: OVOPlanRate
    unit_rates: list[OVOPlanUnitRate]


@dataclass
class OVOPlanGas:
    """Plan gas model."""

    name: str
    exit_fee: OVOPlanRate
    contract_start_date: str
    contract_end_date: Any
    contract_type: str
    is_in_renewal: bool
    has_future_contracts: bool
    mpxn: str
    msn: str
    personal_projection: float
    standing_charge: OVOPlanRate
    unit_rates: list[OVOPlanUnitRate]


@dataclass
class OVOPlans:
    """Plan model."""

    electricity: list[OVOPlanElectricity]
    gas: list[OVOPlanGas] | None



"""Tests for the client module."""

from datetime import datetime, timedelta

from aioresponses import aioresponses
import pytest
from syrupy.assertion import SnapshotAssertion

from ovoenergy import OVOEnergy
from ovoenergy.exceptions import (
    OVOEnergyAPINoCookies,
    OVOEnergyAPINotAuthorized,
    OVOEnergyNoAccount,
)

from . import ACCOUNT, ACCOUNT_BAD, PASSWORD, RESPONSE_JSON_AUTH, USERNAME


@pytest.mark.asyncio
async def test_authorize(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test authorize."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    assert not ovoenergy_client.oauth_expired

    assert ovoenergy_client.oauth is not None

    assert ovoenergy_client.oauth.access_token == snapshot(
        name="authorize_oauth_access_token",
    )
    assert ovoenergy_client.oauth.expires_in == snapshot(
        name="authorize_oauth_expires_in",
    )
    assert ovoenergy_client.oauth.refresh_expires_in == snapshot(
        name="authorize_oauth_refresh_expires_in",
    )
    assert ovoenergy_client.account_id == snapshot(
        name="authorize_account_id",
    )
    assert ovoenergy_client.account_ids == snapshot(
        name="authorize_account_id",
    )
    assert ovoenergy_client.customer_id == snapshot(
        name="authorize_customer_id",
    )
    assert ovoenergy_client.username == snapshot(
        name="authorize_username",
    )


@pytest.mark.asyncio
async def test_bootstrap(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test bootstrap."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    assert await ovoenergy_client.bootstrap_accounts() == snapshot(
        name="bootstrap_accounts",
    )


@pytest.mark.asyncio
async def test_get_daily_usage(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test get daily usage."""
    with pytest.raises(OVOEnergyNoAccount):
        await ovoenergy_client.get_daily_usage("2024-01")

    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    assert await ovoenergy_client.get_daily_usage("2024-01") == snapshot(
        name="daily_usage",
    )


@pytest.mark.asyncio
async def test_get_half_hourly_usage(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test get half hourly usage."""
    with pytest.raises(OVOEnergyNoAccount):
        await ovoenergy_client.get_half_hourly_usage("2024-01-01")

    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    assert await ovoenergy_client.get_half_hourly_usage("2024-01-01") == snapshot(
        name="half_hourly_usage",
    )


@pytest.mark.asyncio
async def test_get_plans(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test get plans."""
    with pytest.raises(OVOEnergyNoAccount):
        await ovoenergy_client.get_plans()

    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    assert await ovoenergy_client.get_plans() == snapshot(
        name="plans",
    )


@pytest.mark.asyncio
async def test_get_footprint(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test get footprint."""
    with pytest.raises(OVOEnergyNoAccount):
        await ovoenergy_client.get_footprint()

    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    assert await ovoenergy_client.get_footprint() == snapshot(
        name="footprint",
    )


@pytest.mark.asyncio
async def test_get_carbon_intensity(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test get carbon intensity."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    assert await ovoenergy_client.get_carbon_intensity() == snapshot(
        name="carbon_intensity",
    )


@pytest.mark.asyncio
async def test_bootstrap_custom_account(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test bootstrap custom account."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    ovoenergy_client.custom_account_id = ACCOUNT

    await ovoenergy_client.bootstrap_accounts()

    assert await ovoenergy_client.get_daily_usage("2024-01") == snapshot(
        name="bootstrap_accounts_custom_account",
    )


@pytest.mark.asyncio
async def test_bad_account(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test bad account."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    ovoenergy_client.custom_account_id = ACCOUNT_BAD

    with pytest.raises(OVOEnergyNoAccount):
        await ovoenergy_client.get_daily_usage("2024-01")


# pylint: disable=protected-access
@pytest.mark.asyncio
async def test_no_cookies(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test no cookies."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    ovoenergy_client._cookies = None

    with pytest.raises(OVOEnergyAPINoCookies):
        await ovoenergy_client.get_daily_usage("2024-01")


# pylint: disable=protected-access
@pytest.mark.asyncio
async def test_no_auth(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test no auth."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    ovoenergy_client._oauth = None

    with pytest.raises(OVOEnergyAPINotAuthorized):
        await ovoenergy_client.get_daily_usage("2024-01")


# pylint: disable=protected-access
@pytest.mark.asyncio
async def test_oauth_expired(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test oauth expired."""
    an_hour_ago = datetime.now() - timedelta(hours=1)

    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    assert ovoenergy_client._oauth is not None

    ovoenergy_client._oauth.expires_at = an_hour_ago

    await ovoenergy_client.get_daily_usage("2024-01")

    ovoenergy_client._oauth.expires_at = an_hour_ago

    mock_aioresponse.clear()
    mock_aioresponse.get(
        "https://my.ovoenergy.com/api/v2/auth/token",
        status=403,
        repeat=True,
    )

    with pytest.raises(OVOEnergyAPINotAuthorized):
        await ovoenergy_client.get_daily_usage("2024-01")


@pytest.mark.asyncio
async def test_forbidden(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test forbidden."""
    await ovoenergy_client.authenticate(USERNAME, PASSWORD)

    await ovoenergy_client.bootstrap_accounts()

    mock_aioresponse.clear()
    mock_aioresponse.get(
        f"https://smartpaymapi.ovoenergy.com/usage/api/daily/{ACCOUNT}?date=2024-01",
        status=403,
        repeat=True,
    )

    with pytest.raises(OVOEnergyAPINotAuthorized):
        await ovoenergy_client.get_daily_usage("2024-01")


@pytest.mark.asyncio
async def test_auth_not_found(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test auth endpoint not found."""
    mock_aioresponse.clear()
    mock_aioresponse.post(
        "https://my.ovoenergy.com/api/v2/auth/login",
        status=404,
        repeat=True,
    )

    assert not await ovoenergy_client.authenticate(USERNAME, PASSWORD)


@pytest.mark.asyncio
async def test_auth_code_not_found(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test auth endpoint code not found."""
    mock_aioresponse.clear()
    mock_aioresponse.post(
        "https://my.ovoenergy.com/api/v2/auth/login",
        status=204,
        repeat=True,
    )

    assert not await ovoenergy_client.authenticate(USERNAME, PASSWORD)


@pytest.mark.asyncio
async def test_auth_code_unknown(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test auth token not found."""
    mock_aioresponse.clear()
    mock_aioresponse.post(
        "https://my.ovoenergy.com/api/v2/auth/login",
        status=200,
        payload={"code": "Unknown"},
        repeat=True,
    )

    assert not await ovoenergy_client.authenticate(USERNAME, PASSWORD)


@pytest.mark.asyncio
async def test_auth_token_not_found(
    ovoenergy_client: OVOEnergy,
    mock_aioresponse: aioresponses,
) -> None:
    """Test auth token not found."""
    mock_aioresponse.clear()
    mock_aioresponse.post(
        "https://my.ovoenergy.com/api/v2/auth/login",
        payload=RESPONSE_JSON_AUTH,
        status=200,
        repeat=True,
    )
    mock_aioresponse.get(
        "https://my.ovoenergy.com/api/v2/auth/token",
        status=404,
        repeat=True,
    )

    assert not await ovoenergy_client.authenticate(USERNAME, PASSWORD)



"""Footprint Models."""

from dataclasses import dataclass
from typing import Any


@dataclass
class OVOFootprintElectricity:
    """Electricity footprint model."""

    carbon_kg: float
    carbon_saved_kg: float
    k_wh: float


@dataclass
class OVOFootprintGas:
    """Gas footprint model."""

    carbon_kg: float
    carbon_saved_kg: float
    k_wh: float


@dataclass
class OVOFootprintBreakdown:
    """Footprint breakdown model."""

    electricity: OVOFootprintElectricity
    gas: OVOFootprintGas


@dataclass
class OVOCarbonFootprint:
    """Carbon footprint model."""

    carbon_kg: float
    carbon_saved_kg: float
    k_wh: float
    breakdown: OVOFootprintBreakdown


@dataclass
class OVOFootprint:
    """Footprint model."""

    from_: str | None
    to: str | None
    carbon_reduction_product_ids: list[Any]
    carbon_footprint: OVOCarbonFootprint | None



"""Footprint Models."""

from dataclasses import dataclass
from typing import Any


@dataclass
class OVOCarbonIntensityForecast:
    """Carbon intensity forecast model."""

    time_from: str
    intensity: float
    level: str
    colour: str
    colour_v2: str


@dataclass
class OVOCarbonIntensity:
    """Carbon intensity model."""

    forecast: list[OVOCarbonIntensityForecast]
    current: str | None
    greentime: Any | None



"""Get energy data from OVO's API."""

import contextlib
from datetime import datetime, timedelta
from http.cookies import SimpleCookie
import logging
from typing import Literal
from uuid import UUID

import aiohttp

from .exceptions import (
    OVOEnergyAPINoCookies,
    OVOEnergyAPINotAuthorized,
    OVOEnergyNoAccount,
)
from .models import (
    OVOCost,
    OVODailyElectricity,
    OVODailyGas,
    OVODailyUsage,
    OVOHalfHour,
    OVOHalfHourUsage,
    OVOInterval,
    OVOMeterReadings,
)
from .models.accounts import Account, BootstrapAccounts, Supply, SupplyPointInfo
from .models.carbon_intensity import OVOCarbonIntensity, OVOCarbonIntensityForecast
from .models.footprint import (
    OVOCarbonFootprint,
    OVOFootprint,
    OVOFootprintBreakdown,
    OVOFootprintElectricity,
    OVOFootprintGas,
)
from .models.oauth import OAuth
from .models.plan import (
    OVOPlanElectricity,
    OVOPlanGas,
    OVOPlanRate,
    OVOPlans,
    OVOPlanUnitRate,
)

_LOGGER = logging.getLogger(__name__)


class OVOEnergy:
    """Class for OVOEnergy."""

    custom_account_id: int | None = None

    def __init__(
        self,
        client_session: aiohttp.ClientSession,
    ) -> None:
        """Initilalize."""
        self._client_session = client_session

        self._bootstrap_accounts: BootstrapAccounts | None = None
        self._cookies: SimpleCookie | None = None
        self._oauth: OAuth | None = None
        self._username: str | None = None

    @property
    def account_id(self) -> int | None:
        """Return account id."""
        if (
            self.custom_account_id is not None
            and self.account_ids is not None
            and self.custom_account_id not in set(self.account_ids)
        ):
            raise OVOEnergyNoAccount("Custom account not found in accounts")

        return (
            self.custom_account_id
            if self.custom_account_id is not None
            else self._bootstrap_accounts.selected_account_id
            if self._bootstrap_accounts
            else None
        )

    @property
    def account_ids(self) -> list[int] | None:
        """Return account ids."""
        return (
            self._bootstrap_accounts.account_ids if self._bootstrap_accounts else None
        )

    @property
    def customer_id(self) -> UUID | None:
        """Return customer id."""
        return (
            self._bootstrap_accounts.customer_id if self._bootstrap_accounts else None
        )

    @property
    def oauth(self) -> OAuth | None:
        """Return OAuth."""
        return self._oauth

    @property
    def oauth_expired(self) -> bool:
        """Return True if OAuth token has expired."""
        return self.oauth is None or self.oauth.expires_at < datetime.now()

    @property
    def username(self) -> str | None:
        """Return username."""
        return self._username

    async def _request(
        self,
        url: str,
        method: Literal["GET"] | Literal["POST"],
        with_cookies: bool = True,
        with_authorization: bool = True,
        **kwargs,
    ):
        """Request."""
        if with_cookies and self._cookies is None:
            raise OVOEnergyAPINoCookies("No cookies set")
        if with_authorization and self._oauth is None:
            raise OVOEnergyAPINotAuthorized("No OAuth token set")

        if with_authorization and self.oauth_expired:
            _LOGGER.debug("OAuth token expired, refreshing: %s", self.oauth)

            if not await self.get_token() or self.oauth is None:
                raise OVOEnergyAPINotAuthorized("No OAuth token set after refresh")

            _LOGGER.debug("OAuth token refreshed: %s", self.oauth)

        response = await self._client_session.request(
            method,
            url,
            cookies=self._cookies if with_cookies else None,
            headers={
                "Authorization": f"Bearer {self.oauth.access_token}"
                if self.oauth
                else None
            }
            if with_authorization
            else None,
            **kwargs,
        )
        with contextlib.suppress(aiohttp.ClientResponseError):
            response.raise_for_status()

        if with_authorization and response.status in [401, 403]:
            raise OVOEnergyAPINotAuthorized(f"Not authorized: {response.status}")

        return response

    async def authenticate(
        self,
        username: str,
        password: str,
    ) -> bool:
        """Authenticate."""
        response = await self._request(
            "https://my.ovoenergy.com/api/v2/auth/login",
            "POST",
            json={
                "username": username,
                "password": password,
                "rememberMe": True,
            },
            with_cookies=False,
            with_authorization=False,
        )
        if response.status != 200:
            return False

        json_response = await response.json()

        if "code" in json_response and json_response["code"] == "Unknown":
            return False

        self._cookies = response.cookies
        self._username = username

        if not await self.get_token():
            return False

        return True

    async def get_token(self) -> OAuth | Literal[False]:
        """Get token."""
        response = await self._request(
            "https://my.ovoenergy.com/api/v2/auth/token",
            "GET",
            with_authorization=False,
        )
        if response.status != 200:
            return False

        json_response = await response.json()

        self._oauth = OAuth(
            access_token=json_response["accessToken"]["value"],
            expires_in=json_response["expiresIn"],
            refresh_expires_in=json_response["refreshExpiresIn"],
            # Set expires_at to current time plus expiresIn (minutes)
            expires_at=datetime.now() + timedelta(minutes=json_response["expiresIn"]),
        )

        return self._oauth

    async def bootstrap_accounts(self) -> BootstrapAccounts:
        """Bootstrap accounts."""
        response = await self._request(
            "https://smartpaymapi.ovoenergy.com/first-login/api/bootstrap/v2/",
            "GET",
        )
        json_response = await response.json()

        self._bootstrap_accounts = BootstrapAccounts(
            account_ids=json_response["accountIds"],
            customer_id=UUID(json_response["customerId"]),
            selected_account_id=json_response["selectedAccountId"],
            is_first_login=json_response.get("isFirstLogin", None),
            accounts=[
                Account(
                    account_id=account.get("accountId", None),
                    is_payg=account.get("isPayg", None),
                    is_blocked=account.get("isBlocked", None),
                    supplies=[
                        Supply(
                            mpxn=supply["mpxn"],
                            fuel=supply["fuel"],
                            is_onboarding=supply["isOnboarding"],
                            start=datetime.fromisoformat(supply["start"])
                            if supply["start"]
                            else None,
                            is_payg=supply["isPayg"],
                            supply_point_info=SupplyPointInfo(
                                meter_type=supply["supplyPointInfo"]["meterType"],
                                meter_not_found=supply["supplyPointInfo"][
                                    "meterNotFound"
                                ],
                                address=supply["supplyPointInfo"]["address"],
                            )
                            if supply["supplyPointInfo"]
                            else None,
                        )
                        for supply in account["supplies"]
                    ]
                    if "supplies" in account
                    else None
                    if "supplies" in account
                    else None,
                )
                for account in json_response["accounts"]
            ]
            if "accounts" in json_response
            else None,
        )

        return self._bootstrap_accounts

    async def get_daily_usage(
        self,
        date: str,
    ) -> OVODailyUsage:
        """Get daily usage data."""
        if self.account_id is None:
            raise OVOEnergyNoAccount("No account found")

        ovo_usage = OVODailyUsage(
            electricity=None,
            gas=None,
        )

        response = await self._request(
            f"https://smartpaymapi.ovoenergy.com/usage/api/daily/{self.account_id}?date={date}",
            "GET",
        )
        json_response = await response.json()

        if "electricity" in json_response:
            electricity = json_response["electricity"]
            if electricity and "data" in electricity:
                ovo_usage.electricity = []
                for usage in electricity["data"]:
                    if usage is not None:
                        ovo_usage.electricity.append(
                            OVODailyElectricity(
                                consumption=usage.get("consumption", None),
                                interval=OVOInterval(
                                    start=datetime.fromisoformat(
                                        usage["interval"]["start"]
                                    ),
                                    end=datetime.fromisoformat(
                                        usage["interval"]["end"]
                                    ),
                                )
                                if "interval" in usage
                                else None,
                                meter_readings=OVOMeterReadings(
                                    start=usage["meterReadings"]["start"],
                                    end=usage["meterReadings"]["end"],
                                )
                                if "meterReadings" in usage
                                else None,
                                has_half_hour_data=usage.get("hasHalfHourData", None),
                                cost=OVOCost(
                                    amount=usage["cost"]["amount"],
                                    currency_unit=usage["cost"]["currencyUnit"],
                                )
                                if "cost" in usage
                                else None,
                            )
                        )

        if "gas" in json_response:
            gas = json_response["gas"]
            if gas and "data" in gas:
                ovo_usage.gas = []
                for usage in gas["data"]:
                    if usage is not None:
                        ovo_usage.gas.append(
                            OVODailyGas(
                                consumption=usage.get("consumption", None),
                                volume=usage.get("volume", None),
                                interval=OVOInterval(
                                    start=datetime.fromisoformat(
                                        usage["interval"]["start"]
                                    ),
                                    end=datetime.fromisoformat(
                                        usage["interval"]["end"]
                                    ),
                                )
                                if "interval" in usage
                                else None,
                                meter_readings=OVOMeterReadings(
                                    start=usage["meterReadings"]["start"],
                                    end=usage["meterReadings"]["end"],
                                )
                                if "meterReadings" in usage
                                else None,
                                has_half_hour_data=usage.get("hasHalfHourData", None),
                                cost=OVOCost(
                                    amount=usage["cost"]["amount"],
                                    currency_unit=usage["cost"]["currencyUnit"],
                                ),
                            )
                        )

        return ovo_usage

    async def get_half_hourly_usage(
        self,
        date: str,
    ) -> OVOHalfHourUsage:
        """Get half hourly usage data."""
        if self.account_id is None:
            raise OVOEnergyNoAccount("No account found")

        ovo_usage = OVOHalfHourUsage(
            electricity=None,
            gas=None,
        )

        response = await self._request(
            f"https://smartpaymapi.ovoenergy.com/usage/api/half-hourly/{self.account_id}?date={date}",
            "GET",
        )
        json_response = await response.json()

        if "electricity" in json_response:
            electricity = json_response["electricity"]
            if electricity and "data" in electricity:
                ovo_usage.electricity = []
                for usage in electricity["data"]:
                    if usage is not None:
                        ovo_usage.electricity.append(
                            OVOHalfHour(
                                consumption=usage["consumption"],
                                interval=OVOInterval(
                                    start=datetime.fromisoformat(
                                        usage["interval"]["start"]
                                    ),
                                    end=datetime.fromisoformat(
                                        usage["interval"]["end"]
                                    ),
                                ),
                                unit=usage["unit"],
                            )
                        )
        if "gas" in json_response:
            gas = json_response["gas"]
            if gas and "data" in gas:
                ovo_usage.gas = []
                for usage in gas["data"]:
                    if usage is not None:
                        ovo_usage.gas.append(
                            OVOHalfHour(
                                consumption=usage["consumption"],
                                interval=OVOInterval(
                                    start=datetime.fromisoformat(
                                        usage["interval"]["start"]
                                    ),
                                    end=datetime.fromisoformat(
                                        usage["interval"]["end"]
                                    ),
                                ),
                                unit=usage["unit"],
                            )
                        )

        return ovo_usage

    async def get_plans(self) -> OVOPlans:
        """Get plans."""
        if self.account_id is None:
            raise OVOEnergyNoAccount("No account found")

        response = await self._request(
            f"https://smartpaymapi.ovoenergy.com/orex/api/plans/{self.account_id}",
            "GET",
        )
        json_response = await response.json()

        return OVOPlans(
            electricity=[
                OVOPlanElectricity(
                    name=json_response["electricity"]["name"],
                    exit_fee=OVOPlanRate(
                        amount=json_response["electricity"]["exitFee"]["amount"],
                        currency_unit=json_response["electricity"]["exitFee"][
                            "currencyUnit"
                        ],
                    ),
                    contract_start_date=json_response["electricity"][
                        "contractStartDate"
                    ],
                    contract_end_date=json_response["electricity"]["contractEndDate"],
                    contract_type=json_response["electricity"]["contractType"],
                    is_in_renewal=json_response["electricity"]["isInRenewal"],
                    has_future_contracts=json_response["electricity"][
                        "hasFutureContracts"
                    ],
                    mpxn=json_response["electricity"]["mpxn"],
                    msn=json_response["electricity"]["msn"],
                    personal_projection=json_response["electricity"][
                        "personalProjection"
                    ],
                    standing_charge=OVOPlanRate(
                        amount=json_response["electricity"]["standingCharge"]["amount"],
                        currency_unit=json_response["electricity"]["standingCharge"][
                            "currencyUnit"
                        ],
                    ),
                    unit_rates=[
                        OVOPlanUnitRate(
                            name=unit_rate["name"],
                            unit_rate=OVOPlanRate(
                                amount=unit_rate["unitRate"]["amount"],
                                currency_unit=unit_rate["unitRate"]["currencyUnit"],
                            ),
                        )
                        for unit_rate in json_response["electricity"]["unitRates"]
                    ],
                )
                for json_response in json_response["electricity"]
            ],
            gas=[
                OVOPlanGas(
                    name=json_response["gas"]["name"],
                    exit_fee=OVOPlanRate(
                        amount=json_response["gas"]["exitFee"]["amount"],
                        currency_unit=json_response["gas"]["exitFee"]["currencyUnit"],
                    ),
                    contract_start_date=json_response["gas"]["contractStartDate"],
                    contract_end_date=json_response["gas"]["contractEndDate"],
                    contract_type=json_response["gas"]["contractType"],
                    is_in_renewal=json_response["gas"]["isInRenewal"],
                    has_future_contracts=json_response["gas"]["hasFutureContracts"],
                    mpxn=json_response["gas"]["mpxn"],
                    msn=json_response["gas"]["msn"],
                    personal_projection=json_response["gas"]["personalProjection"],
                    standing_charge=OVOPlanRate(
                        amount=json_response["gas"]["standingCharge"]["amount"],
                        currency_unit=json_response["gas"]["standingCharge"][
                            "currencyUnit"
                        ],
                    ),
                    unit_rates=[
                        OVOPlanUnitRate(
                            name=unit_rate["name"],
                            unit_rate=OVOPlanRate(
                                amount=unit_rate["unitRate"]["amount"],
                                currency_unit=unit_rate["unitRate"]["currencyUnit"],
                            ),
                        )
                        for unit_rate in json_response["gas"]["unitRates"]
                    ],
                )
                for json_response in json_response["gas"]
            ]
            if "gas" in json_response
            else None,
        )

    async def get_footprint(self) -> OVOFootprint:
        """Get footprint."""
        if self.account_id is None:
            raise OVOEnergyNoAccount("No account found")

        response = await self._request(
            f"https://smartpaymapi.ovoenergy.com/carbon-api/{self.account_id}/footprint",
            "GET",
        )
        json_response = await response.json()

        return OVOFootprint(
            from_=json_response["from"],
            to=json_response["to"],
            carbon_reduction_product_ids=json_response["carbonReductionProductIds"],
            carbon_footprint=OVOCarbonFootprint(
                carbon_kg=json_response["carbonFootprint"]["carbonKg"],
                carbon_saved_kg=json_response["carbonFootprint"]["carbonSavedKg"],
                k_wh=json_response["carbonFootprint"]["kWh"],
                breakdown=OVOFootprintBreakdown(
                    electricity=OVOFootprintElectricity(
                        carbon_kg=json_response["carbonFootprint"]["breakdown"][
                            "electricity"
                        ]["carbonKg"],
                        carbon_saved_kg=json_response["carbonFootprint"]["breakdown"][
                            "electricity"
                        ]["carbonSavedKg"],
                        k_wh=json_response["carbonFootprint"]["breakdown"][
                            "electricity"
                        ]["kWh"],
                    ),
                    gas=OVOFootprintGas(
                        carbon_kg=json_response["carbonFootprint"]["breakdown"]["gas"][
                            "carbonKg"
                        ],
                        carbon_saved_kg=json_response["carbonFootprint"]["breakdown"][
                            "gas"
                        ]["carbonSavedKg"],
                        k_wh=json_response["carbonFootprint"]["breakdown"]["gas"][
                            "kWh"
                        ],
                    ),
                ),
            ),
        )

    async def get_carbon_intensity(self):
        """Get carbon intensity."""
        response = await self._request(
            "https://smartpaymapi.ovoenergy.com/carbon-bff/carbonintensity",
            "GET",
        )
        json_response = await response.json()

        return OVOCarbonIntensity(
            forecast=[
                OVOCarbonIntensityForecast(
                    time_from=forecast["from"],
                    intensity=forecast["intensity"],
                    level=forecast["level"],
                    colour=forecast["colour"],
                    colour_v2=forecast["colourV2"],
                )
                for forecast in json_response["forecast"]
            ],
            current=json_response["current"],
            greentime=json_response["greentime"],
        )



"""Dataclasses for the bootstrap/accounts endpoint."""

from dataclasses import dataclass
from datetime import datetime
from uuid import UUID


@dataclass
class SupplyPointInfo:
    """Supply point info model."""

    meter_type: str | None = None
    meter_not_found: bool | None = None
    address: list[str] | None = None


@dataclass
class Supply:
    """Supply model."""

    mpxn: str | None
    fuel: str | None
    is_onboarding: bool | None
    start: datetime | None
    is_payg: bool | None
    supply_point_info: SupplyPointInfo | None


@dataclass
class Account:
    """Account model."""

    account_id: int
    is_payg: bool | None
    is_blocked: bool | None
    supplies: list[Supply] | None


@dataclass
class BootstrapAccounts:
    """Bootstrap Accounts model."""

    account_ids: list[int]
    customer_id: UUID
    selected_account_id: int
    accounts: list[Account] | None
    is_first_login: bool | None



"""Main."""

import asyncio
from dataclasses import asdict
from datetime import datetime, timedelta

import aiohttp
import typer

from . import OVOEnergy
from ._version import __version__

app = typer.Typer()
loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)


async def _setup_client(
    username: str,
    password: str,
    account: int | None = None,
) -> tuple[OVOEnergy, aiohttp.ClientSession]:
    """Set up OVO Energy client."""
    client_session = aiohttp.ClientSession()
    client = OVOEnergy(
        client_session=client_session,
    )

    if not await client.authenticate(username, password):
        typer.secho("Authentication failed", fg=typer.colors.RED)
        raise typer.Abort()

    await client.bootstrap_accounts()

    if account is not None:
        client.custom_account_id = account

    return (client, client_session)


@app.command(name="bootstrap", short_help="Bootstrap OVO Energy")
def bootstrap(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
) -> None:
    """Authenticate with OVO Energy."""
    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password)
    )

    bootstrap_accounts = loop.run_until_complete(client.bootstrap_accounts())

    typer.secho(
        asdict(bootstrap_accounts),
        fg=typer.colors.GREEN,
    )

    loop.run_until_complete(client_session.close())


@app.command(name="daily", short_help="Get daily usage from OVO Energy")
def daily(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
    account: int = typer.Option(
        None, help="OVO Energy account number (default: first account)"
    ),
    date: str = typer.Option(
        None, help="Date to retrieve data for (default: this month)"
    ),
) -> None:
    """Get daily usage from OVO Energy."""
    if date is None:
        # Get this month
        date = datetime.now().strftime("%Y-%m")

    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password, account)
    )
    ovo_usage = loop.run_until_complete(client.get_daily_usage(date))

    typer.secho(
        asdict(ovo_usage) if ovo_usage is not None else '{"message": "No data"}',
        fg=typer.colors.GREEN,
    )
    loop.run_until_complete(client_session.close())


@app.command(name="halfhourly", short_help="Get half hourly usage from OVO Energy")
def half_hourly(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
    account: int = typer.Option(
        None, help="OVO Energy account number (default: first account)"
    ),
    date: str = typer.Option(
        None, help="Date to retrieve data for (default: this month)"
    ),
) -> None:
    """Get half hourly usage from OVO Energy."""
    if date is None:
        # Get yesterday's date
        date = (datetime.now() - timedelta(days=1)).strftime("%Y-%m-%d")

    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password, account)
    )
    ovo_usage = loop.run_until_complete(client.get_half_hourly_usage(date))

    typer.secho(
        asdict(ovo_usage) if ovo_usage is not None else '{"message": "No data"}',
        fg=typer.colors.GREEN,
    )
    loop.run_until_complete(client_session.close())


@app.command(name="plans", short_help="Get plans from OVO Energy")
def plans(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
    account: int = typer.Option(
        None, help="OVO Energy account number (default: first account)"
    ),
) -> None:
    """Get rates from OVO Energy."""
    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password, account)
    )

    ovo_plans = loop.run_until_complete(client.get_plans())

    typer.secho(
        asdict(ovo_plans) if ovo_plans is not None else '{"message": "No data"}',
        fg=typer.colors.GREEN,
    )
    loop.run_until_complete(client_session.close())


@app.command(name="carbon-footprint", short_help="Get carbon footprint from OVO Energy")
def carbon_footprint(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
    account: int = typer.Option(
        None, help="OVO Energy account number (default: first account)"
    ),
) -> None:
    """Get carbon footprint from OVO Energy."""
    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password, account)
    )
    ovo_footprint = loop.run_until_complete(client.get_footprint())

    typer.secho(
        asdict(ovo_footprint)
        if ovo_footprint is not None
        else '{"message": "No data"}',
        fg=typer.colors.GREEN,
    )
    loop.run_until_complete(client_session.close())


@app.command(name="carbon-intensity", short_help="Get carbon intensity from OVO Energy")
def carbon_intensity(
    username: str = typer.Option(..., help="OVO Energy username"),
    password: str = typer.Option(..., help="OVO Energy password"),
    account: int = typer.Option(
        None, help="OVO Energy account number (default: first account)"
    ),
) -> None:
    """Get carbon intensity from OVO Energy."""
    [client, client_session] = loop.run_until_complete(
        _setup_client(username, password, account)
    )
    ovo_carbon_intensity = loop.run_until_complete(client.get_carbon_intensity())

    typer.secho(
        asdict(ovo_carbon_intensity)
        if ovo_carbon_intensity is not None
        else '{"message": "No data"}',
        fg=typer.colors.GREEN,
    )
    loop.run_until_complete(client_session.close())


@app.command(name="version", short_help="Module Version")
def version() -> None:
    """Display module version."""
    typer.secho(__version__.public(), fg=typer.colors.CYAN)


if __name__ == "__main__":
    app()
